/*
    FFT Display - browser version 1.0
    13 May, 2015 - updated 2 April, 2019
    Copyright © 2010-2019 Harry Whitfield
    mailto:g6auc@arrl.net
*/

/*jslint bitwise, browser */

/*global Complex */

/*property
    PI, add, length, mul, one, pow, sin, sub, swap
*/

/*
Number.prototype.toBitString = function (nBits) {
    var s = this.toString(2);
    while (s.length < nBits) {
        s = "0" + s;
    }
    return s;
};
*/

var bitReverse = (function () {
    "use strict";
    var rev3 = [
    	0, 4, 2, 6,
        1, 5, 3, 7
    ];
    var rev4 = [
        0, 8, 4, 12, 2, 10, 6, 14,
        1, 9, 5, 13, 3, 11, 7, 15
    ];
    var rev5 = [
        0, 16, 8, 24, 4, 20, 12, 28, 2, 18, 10, 26, 6, 22, 14, 30,
        1, 17, 9, 25, 5, 21, 13, 29, 3, 19, 11, 27, 7, 23, 15, 31
    ];

    return function (i, n) {
        var rev;
        var m;

        switch (n) {
        case 8:
            rev = rev3[i];
            break;
        case 16:
            rev = rev4[i];
            break;
        case 32:
            rev = rev5[i];
            break;
        case 64:
            rev = rev3[i & 7] << 3 | rev3[i >> 3];
            break;
        case 128:
            rev = rev4[i & 15] << 3 | rev3[i >> 4];
            break;
        case 256:
            rev = rev4[i & 15] << 4 | rev4[i >> 4];
            break;
        case 512:
            rev = rev5[i & 31] << 4 | rev4[i >> 5];
            break;
        case 1024:
            rev = rev5[i & 31] << 5 | rev5[i >> 5];
            break;
        default:
            rev = 0;
            m = n >> 1;
            while (i !== 0) {
                if ((i & 1) === 1) {
                    rev |= m;
                }
                m >>= 1;
                i >>= 1;
            }
        }
        return rev;
    };
}());

/*
for (var i = 0; i < 16; i += 1) {
    print(i + " " + i.toBitString(4) + " " + bitReverse(i, 16).toBitString(4));
}
*/

function fft(data, nBits, inverse) {    // data is array of complex numbers
    "use strict";
    var step;
	var istep;
	var theta;
	var sinHalfTheta;
	var w0;
	var w;
	var m;
	var prod;
    var n = data.length;
    var sign = (
    	inverse
        ? -1
        : 1
    );
    var i;
    var j;
    var rev;

    if (Math.pow(2, nBits) !== n) {
        throw new Error("Invalid parameters.");
    }

    i = 0;
    while (i < n) {
        rev = bitReverse(i, n);
        if (rev > i) {
            data[i].swap(data[rev]);
        }
        i += 1;
    }

    step = 1;
    while (step < n) {
        istep = step << 1;
        theta = sign * Math.PI / step;
        sinHalfTheta = Math.sin(0.5 * theta);
        w0 = new Complex(-2 * sinHalfTheta * sinHalfTheta, Math.sin(theta));
        w = Complex.one;

        m = 0;
        while (m < step) {
            i = m;
            while (i < n) {
                j = i + step;
                prod = w.mul(data[j]);
                data[j] = data[i].sub(prod);
                data[i] = data[i].add(prod);
                i += istep;
            }
            w = w.add(w.mul(w0));
            m += 1;
        }
        step <<= 1;
    }
}
